import { Component, ElementRef, OnInit, ViewChild } from '@angular/core';
import { MatSnackBar } from '@angular/material/snack-bar';
import { EnvMonitorService } from '../services/env-monitor.service';
import {COMMA, ENTER, P} from '@angular/cdk/keycodes';
import { MatAutocompleteSelectedEvent, MatAutocompleteTrigger } from '@angular/material/autocomplete';
import { DeviceData, DeviceSettings } from '../data-model.interface';
import { commonOptions, tempOptions, humidOptions, vocIndexOptions, co2ppmOptions } from './chart-options';
import { ApexAxisChartSeries } from 'ng-apexcharts';

@Component({
  selector: 'app-prehled',
  templateUrl: './prehled.component.html',
  styleUrls: ['./prehled.component.scss']
})
export class PrehledComponent implements OnInit {

  deviceIdList: string[] = [];
  deviceIdFilteredList: string[] = [];
  deviceIdSelectedList: string[] = [];

  deviceSettingsList: DeviceSettings[] = [];

  deviceIdAutocomplete: string = null;
  separatorKeysCodes: number[] = [ENTER, COMMA];

  public tempSeries: ApexAxisChartSeries = [];
  public humidSeries: ApexAxisChartSeries = [];
  public vocIndexSeries: ApexAxisChartSeries = [];
  public co2ppmSeries: ApexAxisChartSeries = [];

  public commonOptions = commonOptions;
  public tempOptions = tempOptions;
  public humidOptions = humidOptions;
  public vocIndexOptions = vocIndexOptions;
  public co2ppmOptions = co2ppmOptions;

  @ViewChild('deviceIdAutoInputCmp') deviceIdAutoInputCmp: ElementRef<HTMLInputElement>;
  @ViewChild(MatAutocompleteTrigger) autocomplete: MatAutocompleteTrigger;

  constructor(
    private envMonitorService: EnvMonitorService,
    private snackBar: MatSnackBar
  ) {}

  ngOnInit() {
    this.loadDeviceIdList();
  }

  /**
   * Načtení listu id zařízení
   */
  private loadDeviceIdList() {
    this.envMonitorService.getSettings().subscribe(
      (res) => {
        this.emptyLists();

        this.deviceIdList = res.map(item => {
          return item.payload.doc.id;
        });

        this.deviceSettingsList = res.map(item => {
          return {
            id: item.payload.doc.id,
            ...item.payload.doc.data() as DeviceSettings
          }
        });

        if (this.deviceIdList.length > 0) {
          this.deviceIdSelectedList.push(this.deviceIdList[0]);

          this.loadMeasuredData(this.deviceIdSelectedList[0]);
        }

        this.refreshDeviceFilteredList();
      },
      () => {
        this.snackBar.open('Při načítání dat došlo k chybě.', 'Zavřít', {panelClass: 'err-snackbar'});
      }
    );
  }

  /**
   * Načte data daného zařízení z firestore
   * @param deviceId ID zařízení
   */
  private loadMeasuredData(deviceId: string) {

    // TODO - pocatecni zoom grafu
    this.envMonitorService.getMeasuredData(deviceId).subscribe(
      (res) => {
        // Načtení dat pouze v případě, že ještě nejsou načtena
        if (this.tempSeries.findIndex(item => item.name === deviceId) === -1) {
          const deviceData = res.map(item => {
            const data = item.payload.doc.data() as any;
            data.timestamp = this.envMonitorService.convertTimestampToDate(data.timestamp);

            return {
              id: item.payload.doc.id,
              ...data as DeviceData
            }
          });

          this.addSeriesData(deviceId, deviceData);
        }
      },
      () => {
        this.snackBar.open('Při načítání dat došlo k chybě.', 'Zavřít', {panelClass: 'err-snackbar'});
      }
    );
  }

  /**
   * Ošetření opětovného dotáhnutí dat
   */
  private emptyLists() {
    this.deviceIdList = [];
    this.deviceIdFilteredList = [];
    this.deviceIdSelectedList = [];
  }

  /**
   * Vyfiltruje položky pro autocomplete
   * @param filter 
   */
  filterDeviceId(filter: string) {
    if (!filter) {
      this.refreshDeviceFilteredList();
      return;
    }

    this.deviceIdFilteredList = this.deviceIdList.filter(item =>
      !this.deviceIdSelectedList.includes(item)
      && item.toLocaleLowerCase().includes(filter.toLocaleLowerCase())
    );
  }

  /**
   * Obsluha události odebrání vybraného zařízení
   * @param event 
   */
  deviceIdRemoved(event: string) {
    const index = this.deviceIdSelectedList.findIndex(item => item === event);
    this.deviceIdSelectedList.splice(index, 1);

    // Workaround pro správné fungování zobrazení
    // Náhled grafů se updatuje jen při updatu reference dat
    const tempSeriesNew = Object.assign([], this.tempSeries);
    const tempSeriesIndex = tempSeriesNew.findIndex(item => item.name === event);
    if (tempSeriesIndex !== -1) {
      tempSeriesNew.splice(tempSeriesIndex, 1);
      this.tempSeries = tempSeriesNew;
    }

    const humidSeriesNew = Object.assign([], this.humidSeries);
    const humidSeriesIndex = humidSeriesNew.findIndex(item => item.name === event);
    if (humidSeriesIndex !== -1) {
      humidSeriesNew.splice(humidSeriesIndex, 1);
      this.humidSeries = humidSeriesNew;
    }

    const vocIndexSeriesNew = Object.assign([], this.vocIndexSeries);
    const vocIndex = vocIndexSeriesNew.findIndex(item => item.name === event);
    if (vocIndex !== -1) {
      vocIndexSeriesNew.splice(vocIndex, 1);
      this.vocIndexSeries = vocIndexSeriesNew;
    }

    const co2ppmSeriesNew = Object.assign([], this.co2ppmSeries);
    const co2ppmIndex = co2ppmSeriesNew.findIndex(item => item.name == event);
    if (co2ppmIndex !== -1) {
      co2ppmSeriesNew.splice(co2ppmIndex, 1);
      this.co2ppmSeries = co2ppmSeriesNew;
    }

    this.refreshDeviceFilteredList();
  }

  /**
   * Obsluha události vybrání položky v autocomplete
   * @param event 
   */
  deviceAutocompleteSelected(event: MatAutocompleteSelectedEvent) {
    if (!event?.option?.value) {
      return;
    }

    const selected = event.option.value;

    this.deviceIdSelectedList.push(selected);

    this.deviceIdAutocomplete = '';
    this.deviceIdAutoInputCmp.nativeElement.value = '';

    this.loadMeasuredData(selected);

    this.refreshDeviceFilteredList();

    this.reopenAutocomplete();
  }

  /**
   * Automaticky otevře autocomplete hned po výběru
   */
   reopenAutocomplete() {
    // Workaround pro automatické rozbalení nového autocomplete hned po výběru
    // Musí být takto, jinak se hned opět zavře v reakci na výběr
    setTimeout(() => {
      this.autocomplete.openPanel();
    }, 100);
  }

  /**
   * Aktualizuje list pro autocomplete.
   * Do listu se přidají pouze zařízení, která ještě nejsou vybrána
   */
  private refreshDeviceFilteredList() {
    this.deviceIdFilteredList = [];

    this.deviceIdList.forEach(item => {
      if (!this.deviceIdSelectedList.includes(item)) {
        this.deviceIdFilteredList.push(item);
      }
    });
  }

  /**
   * Přidá příchozí data do listů pro grafy
   * @param deviceId ID zařízení
   * @param deviceDataList Příchozí data
   */
  addSeriesData(deviceId: string, deviceDataList: DeviceData[]) {
    let tempData = [];
    let humidData = [];
    let vocIndexData = [];
    let co2ppmData = [];

    deviceDataList.forEach(item => {
      tempData.push([item.timestamp, item.temp]);
      humidData.push([item.timestamp, item.humid]);
      if ('vocIndex' in item) {
        vocIndexData.push([item.timestamp, item.vocIndex]);
      }
      if ('co2ppm' in item) {
        co2ppmData.push([item.timestamp, item.co2ppm]);
      }
    });

    // Workaround pro správné fungování zobrazení
    // Náhled grafů se updatuje jen při updatu reference dat
    const tempSeriesNew = Object.assign([], this.tempSeries);
    tempSeriesNew.push({name: deviceId, data: tempData});
    this.tempSeries = tempSeriesNew;

    const humidSeriesNew = Object.assign([], this.humidSeries);
    humidSeriesNew.push({name: deviceId, data: humidData});
    this.humidSeries = humidSeriesNew;

    const vocIndexSeriesNew = Object.assign([], this.vocIndexSeries);
    vocIndexSeriesNew.push({name: deviceId, data: vocIndexData});
    this.vocIndexSeries = vocIndexSeriesNew;

    const co2ppmSeriesNew = Object.assign([], this.co2ppmSeries);
    co2ppmSeriesNew.push({name: deviceId, data: co2ppmData});
    this.co2ppmSeries = co2ppmSeriesNew;
  }

}
