#include "stdlib.h"
#include "string.h"
#include "esp_sleep.h"
#include "cJSON.h"
#include "esp_log.h"
#include "driver/uart.h"
#include "sys/time.h"

#include "communication.h"
#include "env_monitor.h"
#include "sensors.h"
#include "hdc2080.h"
#include "sgp40.h"
#include "sensirion_voc_algorithm.h"
#include "rgb.h"
#include "scd30.h"
#include "battery_adc.h"

#define WAKEUP_BTN1_GPIO            34
#define WAKEUP_BTN2_GPIO            35
#define DATA_BUFF_LENGTH            3500
#define BATT_MEASURE_TIME           60*60*4         // Period of battery measurement in seconds
#define MEASURE_VOC_MASK            0x02
#define MEASURE_CO2_MASK            0x01
#define VOC_ALGORITHM_TIME          1               // Period of VOC algorithm in seconds
#define VOC_ALGORITHM_HDC2080_TIME  20              // Period of measuring temperature and humidity for VOC alforithm in seconds

#define TAG "ENV_MONITOR"

static RTC_DATA_ATTR bool is_initialized;           // True if start init has been done
static RTC_DATA_ATTR bool batt_state_sync;          // True if battery state is in sync with Firestore
static RTC_DATA_ATTR uint16_t send_time;            // Period of sending data to Firestore in seconds
static RTC_DATA_ATTR uint16_t measure_time;         // Wake up time in seconds
static RTC_DATA_ATTR bool measure_co2;              // True if CO2 measure is required
static RTC_DATA_ATTR bool measure_voc;              // True if VOC index measure is required
static RTC_DATA_ATTR uint8_t voc_algorithm_cnt;     // Counter of VOC algorithm number processed
static RTC_DATA_ATTR bool sntp_retry_next;          // True if previous sync attempt failed
static RTC_DATA_ATTR bool load_settings_retry_next; // True if previous load settings attempt failed

static RTC_DATA_ATTR struct timeval sleep_enter_time;
static RTC_DATA_ATTR struct timeval next_measure_time;
static RTC_DATA_ATTR struct timeval next_send_time;
static RTC_DATA_ATTR struct timeval next_battery_measure_time;
static RTC_DATA_ATTR struct timeval next_voc_algorithm_time;
static RTC_DATA_ATTR uint8_t prev_sntp_sync_hour;

static RTC_DATA_ATTR char mac_id[13];

static RTC_DATA_ATTR uint16_t data_buff_index;
static RTC_DATA_ATTR uint8_t data_buff[DATA_BUFF_LENGTH];

static RTC_DATA_ATTR VocAlgorithmParams voc_algorithm_params;
static int32_t voc_index;
static float co2_ppm;

static long env_monitor_timestamp;

// Tasks to do
static bool measure_task_ticked = false;
static bool send_task_ticked = false;
static bool battery_measure_task_ticked = false;
static bool voc_algorithm_task_ticked = false;

void env_monitor_get_mac_id() {
    uint8_t esp_mac[6];
    esp_efuse_mac_get_default((uint8_t *)esp_mac);
    for (int i = 0; i < sizeof(esp_mac); i++) {
        sprintf(&mac_id[2*i], "%02x", esp_mac[i]);
    }
    printf("MAC ID: %s\n", mac_id);
}

void env_monitor_set_default_settings() {
    if (measure_time == 0) {
        measure_time = 60;
    }
    printf("\nWake up time: %d\n", measure_time);

    if (send_time == 0) {
        send_time = 600;
    }
    printf("\nSend time: %d\n", send_time);

    measure_voc = true;
    measure_co2 = true;

    printf("Init done\n");
}

void env_monitor_voc_init() {
    VocAlgorithm_init(&voc_algorithm_params);
}

void env_monitor_set_default_tasks_time() {
    struct timeval now;
    gettimeofday(&now, NULL);

    next_measure_time.tv_sec = now.tv_sec;
    next_send_time.tv_sec = now.tv_sec;
    next_battery_measure_time.tv_sec = now.tv_sec;
    next_voc_algorithm_time.tv_sec = now.tv_sec;

    struct tm timeinfo;
    gmtime_r(&now.tv_sec, &timeinfo);

    prev_sntp_sync_hour = timeinfo.tm_hour;

    is_initialized = true;
}

void env_monitor_check_tasks() {
    struct timeval now;
    gettimeofday(&now, NULL);

    if (now.tv_sec >= next_measure_time.tv_sec) {
        printf("Measure task ticked\n");
        measure_task_ticked = true;
    }

    if (now.tv_sec >= next_send_time.tv_sec) {
        printf("Send task ticked\n");
        send_task_ticked = true;
    }

    if (now.tv_sec >= next_battery_measure_time.tv_sec) {
        printf("Battery measure task ticked\n");
        battery_measure_task_ticked = true;
    }

    if (measure_voc && (now.tv_sec >= next_voc_algorithm_time.tv_sec)) {
        printf("VOC algorithm task ticked\n");
        voc_algorithm_task_ticked = true;
    }
}

bool env_monitor_measure_task_ticked() {
    return measure_task_ticked;
}

bool env_monitor_send_task_ticked() {
    return send_task_ticked;
}

bool env_monitor_battery_measure_task_ticked() {
    return battery_measure_task_ticked;
}

bool env_monitor_voc_algorithm_task_ticked() {
    return voc_algorithm_task_ticked;
}

bool env_monitor_is_initialized() {
    return is_initialized;
}

bool env_monitor_get_sntp_retry_next() {
    return sntp_retry_next;
}

void env_monitor_set_sntp_retry_next(bool retry_next) {
    sntp_retry_next = retry_next;
}

bool env_monitor_get_load_settings_retry_next() {
    return load_settings_retry_next;
}

void env_monitor_set_load_settings_retry_next(bool retry_next) {
    load_settings_retry_next = retry_next;
}

bool env_monitor_get_sntp_sync_time_flag() {
    bool ret = false;

    struct timeval now;
    gettimeofday(&now, NULL);

    struct tm timeinfo;
    gmtime_r(&now.tv_sec, &timeinfo);

    if ((timeinfo.tm_hour < prev_sntp_sync_hour) || ((timeinfo.tm_hour - prev_sntp_sync_hour) == 6)) {
        ret = true;
        prev_sntp_sync_hour = timeinfo.tm_hour;
        printf("Time to get current time from SNTP server\n");
    }

    return ret;
}

void env_monitor_sleep_start() {
    int sleep_time = env_monitor_count_sleep_time_sec();

    printf("Enabling timer wakeup, %ds\n", sleep_time);
    esp_sleep_enable_timer_wakeup(sleep_time * 1000000);

    const uint64_t ext_wakeup_pin_1_mask = 1ULL << WAKEUP_BTN1_GPIO;
    const uint64_t ext_wakeup_pin_2_mask = 1ULL << WAKEUP_BTN2_GPIO;

    esp_sleep_enable_ext1_wakeup(ext_wakeup_pin_1_mask | ext_wakeup_pin_2_mask, ESP_EXT1_WAKEUP_ANY_HIGH);

    printf("Entering deep sleep\n");
    gettimeofday(&sleep_enter_time, NULL);

    esp_deep_sleep_start();
}

int env_monitor_count_sleep_time_sec() {
    struct timeval now;
    gettimeofday(&now, NULL);

    // Find the first task in the future
    long next_task_time = next_send_time.tv_sec;
    if (next_measure_time.tv_sec < next_task_time) {
        next_task_time = next_measure_time.tv_sec;
    }
    if (measure_voc && (next_voc_algorithm_time.tv_sec < next_task_time)) {
        next_task_time = next_voc_algorithm_time.tv_sec;
    }

    return ((next_task_time >= now.tv_sec) ? (next_task_time - now.tv_sec) : 1);
}

env_monitor_wakeup_t env_monitor_get_wakeup_cause() {
    env_monitor_wakeup_t ret = ENV_MONITOR_WAKEUP_UNDEFINED;

    esp_sleep_wakeup_cause_t cause = esp_sleep_get_wakeup_cause();

    switch (cause) {
        case ESP_SLEEP_WAKEUP_TIMER: {
            printf("Wake up from timer\n");
            ret = ENV_MONITOR_WAKEUP_TIMER;
            break;
        }
        case ESP_SLEEP_WAKEUP_EXT1: {
            uint64_t wakeup_pin_mask = esp_sleep_get_ext1_wakeup_status();
            if (wakeup_pin_mask != 0) {
                int pin = __builtin_ffsll(wakeup_pin_mask) - 1;
                printf("Wake up from GPIO %d\n", pin);

                if (pin == 34) {
                    ret = ENV_MONITOR_WAKEUP_GPIO34;
                }
                else {
                    ret = ENV_MONITOR_WAKEUP_GPIO35;
                }
            } else {
                printf("Wake up from GPIO\n");
            }
            break;
        }
        default:
            printf("Not a deep sleep reset\n");
    }

    return ret;
}

esp_err_t env_monitor_load_settings() {
    esp_err_t ret = ESP_OK;

    char* dev_settings;

    firestore_err_t ret_firestore = com_get_document("settings", mac_id, &dev_settings);

    if (ret_firestore == FIRESTORE_OK) {

        // Parse attributes from json
        cJSON *root = cJSON_Parse(dev_settings);

        if (root != NULL) {
            cJSON *fields = cJSON_GetObjectItem(root, "fields");

            if (fields != NULL) {
                // Parse wakeup time
                cJSON *measure_time_json = cJSON_GetObjectItem(fields, "measureTime");

                if (measure_time_json != NULL) {
                    char *measure_time_str = cJSON_GetObjectItem(measure_time_json, "integerValue")->valuestring;

                    if (measure_time_str != NULL) {
                        uint16_t measure_time_int = atoi(measure_time_str);

                        // Set wakeup time
                        if (measure_time_int != 0) {
                            measure_time = measure_time_int;
                        }
                        else {
                            ret = ESP_FAIL;
                            printf("Measure time conversion failed\n");
                        }
                        printf("Measure time set: %d\n", measure_time);
                    }
                    else {
                        ret = ESP_FAIL;
                    }
                }
                else {
                    ret = ESP_FAIL;
                }

                // Parse send time
                cJSON *send_time_json = cJSON_GetObjectItem(fields, "sendTime");

                if (send_time_json != NULL) {
                    char *send_time_str = cJSON_GetObjectItem(send_time_json, "integerValue")->valuestring;

                    if (send_time_str != NULL) {
                        uint16_t send_time_int = atoi(send_time_str);

                        // Set send time
                        if (send_time_int != 0) {
                            send_time = send_time_int;
                        }
                        else {
                            ret = ESP_FAIL;
                            printf("Send time conversion failed\n");
                        }
                        printf("Send time set: %d\n", send_time);
                    }
                    else {
                        ret = ESP_FAIL;
                    }
                }
                else {
                    ret = ESP_FAIL;
                }

                // Parse measure CO2 flag
                cJSON *measure_co2_json = cJSON_GetObjectItem(fields, "measureCo2");
                
                if (measure_co2_json != NULL) {
                    cJSON *measure_co2_bool_json = cJSON_GetObjectItem(measure_co2_json, "booleanValue");

                    if (measure_co2_bool_json != NULL) {

                        // Set measure CO2 flag
                        measure_co2 = cJSON_IsTrue(measure_co2_bool_json);
                        printf(measure_co2 ? "Measure CO2 true\n" : "Measure CO2 false\n");
                    }
                    else {
                        ret = ESP_FAIL;
                    }
                }
                else {
                    ret = ESP_FAIL;
                }

                // Parse measure VOC flag
                cJSON *measure_voc_json = cJSON_GetObjectItem(fields, "measureVoc");

                if (measure_voc_json != NULL) { 
                    cJSON *measure_voc_bool_json = cJSON_GetObjectItem(measure_voc_json, "booleanValue");

                    if (measure_voc_bool_json != NULL) {

                        // Set measure VOC flag
                        measure_voc = cJSON_IsTrue(measure_voc_bool_json);
                        printf(measure_voc ? "Measure VOC true\n" : "Measure VOC false\n");
                    }
                    else {
                        ret = ESP_FAIL;
                    }
                }
                else {
                    ret = ESP_FAIL;
                }

                struct timeval now;
                gettimeofday(&now, NULL);

                next_send_time.tv_sec = now.tv_sec;
                next_measure_time.tv_sec = now.tv_sec;
                next_voc_algorithm_time.tv_sec = now.tv_sec;

                cJSON_Delete(root);
            }
            else {
                ret = ESP_FAIL;
            }
        }
        else {
            ret = ESP_FAIL;
        }
    }
    else {
        ret = ESP_FAIL;
    }

    return ret;
}

static int16_t env_monitor_measure_voc_index() {
    int16_t ret;

    int32_t humid_sgp40_format = (int32_t)(hdc2080_get_humid()*1000);
    int32_t temp_sgp40_format = (int32_t)(hdc2080_get_temp()*1000);
    uint16_t sraw;

    ret = sgp40_measure_raw_with_rht_blocking_read(humid_sgp40_format, temp_sgp40_format, &sraw);

    if (ret == NO_ERROR) {
        VocAlgorithm_process(&voc_algorithm_params, sraw, &voc_index);
        printf("sraw: %d\n", sraw);
        printf("VOC Index: %d\n", voc_index);
    }
    else {
        printf("voc error\n");
    }

    return ret;
}

int16_t env_monitor_process_voc_algorithm() {
    int16_t ret = NO_ERROR;

    if (!measure_task_ticked) {
        voc_algorithm_cnt = ((voc_algorithm_cnt+1) % VOC_ALGORITHM_HDC2080_TIME);
        printf("VOC algorithm cnt: %d\n", voc_algorithm_cnt);

        if (voc_algorithm_cnt == 0) {
            hdc2080_measure();
            printf("VOC algorithm HDC2080 measure\n");
        }
        ret = env_monitor_measure_voc_index();

        next_voc_algorithm_time.tv_sec += VOC_ALGORITHM_TIME;
    }

    return ret;
}

esp_err_t env_monitor_measure_data() {
    esp_err_t ret;

    // HDC2080
    ret = hdc2080_measure();

    if (ret == ESP_OK) {
        printf("Temperature: %f [°C]\n", hdc2080_get_temp());
        env_monitor_timestamp = next_measure_time.tv_sec;

        if (measure_voc) {
            env_monitor_measure_voc_index();
            voc_algorithm_cnt = 0;
            next_voc_algorithm_time.tv_sec += VOC_ALGORITHM_TIME;
        }
    }

    // SCD30 if required
    if (measure_co2) {
        int16_t scd30_ret = sens_scd30_init();

        if (scd30_ret == NO_ERROR) {
            uint16_t scd30_data_ready = 0;
            float temperature, relative_humidity;
            uint32_t timeout = 0;
            uint32_t interval_in_seconds = 2;

            for(uint16_t i = 0; i < 4; i++) {
                for (timeout = 0; timeout < (interval_in_seconds * 12); ++timeout) {
                    scd30_ret = scd30_get_data_ready(&scd30_data_ready);
                    if (scd30_ret != NO_ERROR) {
                        ESP_LOGE(TAG, "Error reading data_ready flag: %i", scd30_ret);
                    }
                    if (scd30_data_ready) {
                        break;
                    }
                    sensirion_sleep_usec(100000);
                }

                if (!scd30_data_ready) {
                    ESP_LOGE(TAG, "SCD30 timeout waiting for data_ready");
                    //rgb_light_ms(5, 1, 0, 500);
                }
                else {
                    scd30_ret =
                        scd30_read_measurement(&co2_ppm, &temperature, &relative_humidity);
                    if (scd30_ret != NO_ERROR) {
                        ESP_LOGE(TAG, "SCD30 error reading measurement");

                    } else {
                        printf("measured co2 concentration: %0.2f ppm, "
                                "measured temperature: %0.2f degreeCelsius, "
                                "measured humidity: %0.2f %%RH\n",
                                co2_ppm, temperature, relative_humidity);
                    }
                }
            }

            sens_scd30_stop();
        }
    }

    next_measure_time.tv_sec += measure_time;

    rgb_light_ms(0, 0, 5, 500);

    return ret;
}

firestore_err_t env_monitor_send_data() {
    firestore_err_t ret;

    if (data_buff_index > 0) {
        ret = env_monitor_send_data_from_buff();
    }
    else {
        char document_id[30];
        sprintf(document_id, "%ld", env_monitor_timestamp);

        // Create document json object
        cJSON *document_json = cJSON_CreateObject();
        cJSON *fields_json = cJSON_CreateObject();
        cJSON_AddItemToObject(document_json, "fields", fields_json);

        // Add timestamp to json
        char timestamp_str[25];
        com_get_firestore_timestamp_value(env_monitor_timestamp, timestamp_str);
        cJSON *timestamp_json = cJSON_CreateObject();
        cJSON_AddStringToObject(timestamp_json, "timestampValue", timestamp_str);
        cJSON_AddItemToObject(fields_json, "timestamp", timestamp_json);

        // Add temperature to json
        cJSON *temp_json = cJSON_CreateObject();
        cJSON_AddNumberToObject(temp_json, "doubleValue", hdc2080_get_temp());
        cJSON_AddItemToObject(fields_json, "temp", temp_json);

        // Add humidity to json
        cJSON *humid_json = cJSON_CreateObject();
        cJSON_AddNumberToObject(humid_json, "doubleValue", hdc2080_get_humid());
        cJSON_AddItemToObject(fields_json, "humid", humid_json);

        // Add VOC index to json if required
        if (measure_voc) {
            cJSON *voc_index_json = cJSON_CreateObject();
            cJSON_AddNumberToObject(voc_index_json, "integerValue", voc_index);
            cJSON_AddItemToObject(fields_json, "vocIndex", voc_index_json);
        }

        // Add CO2 to json if required
        if (measure_co2) {
            cJSON *co2_ppm_json = cJSON_CreateObject();
            cJSON_AddNumberToObject(co2_ppm_json, "doubleValue", co2_ppm);
            cJSON_AddItemToObject(fields_json, "co2ppm", co2_ppm_json);
        }

        char *document = cJSON_Print(document_json);

        ret = com_add_document(mac_id, document_id, document);

        cJSON_Delete(document_json);
        cJSON_free(document);
    }

    if (measure_task_ticked && (ret != FIRESTORE_OK)) {
        env_monitor_add_data_to_buff();
    }

    next_send_time.tv_sec += send_time;

    return ret;
}

firestore_err_t env_monitor_send_data_from_buff() {
    firestore_err_t ret = FIRESTORE_OK;

    // Create document json object
    cJSON *document_json = cJSON_CreateObject();
    cJSON *fields_json = cJSON_CreateObject();
    cJSON_AddItemToObject(document_json, "fields", fields_json);

    cJSON *id_json = cJSON_CreateObject();
    cJSON_AddStringToObject(id_json, "stringValue", mac_id);
    cJSON_AddItemToObject(fields_json, "deviceId", id_json);

    cJSON *data_json = cJSON_CreateObject();
    cJSON_AddItemToObject(fields_json, "data", data_json);

    cJSON *array_value_json = cJSON_CreateObject();
    cJSON_AddItemToObject(data_json, "arrayValue", array_value_json);

    cJSON *values_json = cJSON_AddArrayToObject(array_value_json, "values");

    uint16_t record_index = 0;
    uint16_t byte_index = 0;
    long timestamp_value;
    float temp_value;
    float humid_value;
    int32_t voc_index_value;
    float co2_ppm_value;
    char timestamp_str[30];

    cJSON *value, *map_value_json, *value_fields_json;
    cJSON *timestamp_json, *temp_json, *humid_json, *voc_index_json, *co2_ppm_json;

    while (record_index <= data_buff_index) {
        if (data_buff[record_index] == 0x0) {
			break;
    	}

        value = cJSON_CreateObject();
        
        map_value_json = cJSON_CreateObject();
        cJSON_AddItemToObject(value, "mapValue", map_value_json);

        value_fields_json = cJSON_CreateObject();
        cJSON_AddItemToObject(map_value_json, "fields", value_fields_json);

        byte_index = record_index + 1;

        // Add timestamp to json
        timestamp_value = *(long *)(data_buff + byte_index);
        printf("Measure timestamp from buff: %ld\n", timestamp_value);
        byte_index += sizeof(timestamp_value);

        com_get_firestore_timestamp_value(timestamp_value, timestamp_str);
        timestamp_json = cJSON_CreateObject();
        cJSON_AddStringToObject(timestamp_json, "timestampValue", timestamp_str);
        cJSON_AddItemToObject(value_fields_json, "timestamp", timestamp_json);

        // Add temperature to json
        temp_value = *(float *)(data_buff + byte_index);
        byte_index += sizeof(temp_value);

        temp_json = cJSON_CreateObject();
        cJSON_AddNumberToObject(temp_json, "doubleValue", temp_value);
        cJSON_AddItemToObject(value_fields_json, "temp", temp_json);

        // Add humidity to json
        humid_value = *(float *)(data_buff + byte_index);
        byte_index += sizeof(humid_value);

        humid_json = cJSON_CreateObject();
        cJSON_AddNumberToObject(humid_json, "doubleValue", humid_value);
        cJSON_AddItemToObject(value_fields_json, "humid", humid_json);

        // Add VOC index to json if required
        if (data_buff[record_index] & MEASURE_VOC_MASK) {
            voc_index_value = *(int32_t *)(data_buff + byte_index);
            byte_index += sizeof(voc_index_value);

            voc_index_json = cJSON_CreateObject();
            cJSON_AddNumberToObject(voc_index_json, "integerValue", voc_index_value);
            cJSON_AddItemToObject(value_fields_json, "vocIndex", voc_index_json);
        }

        // Add CO2 to json if required
        if (data_buff[record_index] & MEASURE_CO2_MASK) {
            co2_ppm_value = *(float *)(data_buff + byte_index);
            byte_index += sizeof(co2_ppm_value);

            co2_ppm_json = cJSON_CreateObject();
            cJSON_AddNumberToObject(co2_ppm_json, "doubleValue", co2_ppm_value);
            cJSON_AddItemToObject(value_fields_json, "co2ppm", co2_ppm_json);
        }

        // Add data to array
        cJSON_AddItemToArray(values_json, value);

        record_index = byte_index;
    };

    // Add new data to json if measure task ticked
    if (measure_task_ticked) {
        printf("measure_task_ticked\n");
        value = cJSON_CreateObject();

        map_value_json = cJSON_CreateObject();
        cJSON_AddItemToObject(value, "mapValue", map_value_json);

        value_fields_json = cJSON_CreateObject();
        cJSON_AddItemToObject(map_value_json, "fields", value_fields_json);

        // Add timestamp to json
        com_get_firestore_timestamp_value(env_monitor_timestamp, timestamp_str);
        timestamp_json = cJSON_CreateObject();
        printf("Measure timestamp not from buff: %ld\n", env_monitor_timestamp);
        cJSON_AddStringToObject(timestamp_json, "timestampValue", timestamp_str);
        cJSON_AddItemToObject(value_fields_json, "timestamp", timestamp_json);

        // Add temperature to json
        temp_json = cJSON_CreateObject();
        cJSON_AddNumberToObject(temp_json, "doubleValue", hdc2080_get_temp());
        cJSON_AddItemToObject(value_fields_json, "temp", temp_json);

        // Add humidity to json
        humid_json = cJSON_CreateObject();
        cJSON_AddNumberToObject(humid_json, "doubleValue", hdc2080_get_humid());
        cJSON_AddItemToObject(value_fields_json, "humid", humid_json);

        // Add VOC index to json if required
        if (measure_voc) {
            voc_index_json = cJSON_CreateObject();
            cJSON_AddNumberToObject(voc_index_json, "integerValue", voc_index);
            cJSON_AddItemToObject(value_fields_json, "vocIndex", voc_index_json);
        }

        // Add CO2 to json if required
        if (measure_co2) {
            co2_ppm_json = cJSON_CreateObject();
            cJSON_AddNumberToObject(co2_ppm_json, "doubleValue", co2_ppm);
            cJSON_AddItemToObject(value_fields_json, "co2ppm", co2_ppm_json);
        }

        // Add data to array
        cJSON_AddItemToArray(values_json, value);
    }

    char *document = cJSON_Print(document_json);

    char document_id[30];
    sprintf(document_id, "%s_%ld", mac_id, next_send_time.tv_sec);

    ret = com_add_document("bufferRecords", document_id, document);

    cJSON_Delete(document_json);
    cJSON_free(document);

    if (ret == FIRESTORE_OK) {
        data_buff_index = 0;
    }

    return ret;
}

void env_monitor_add_data_to_buff() {
    if (data_buff_index < DATA_BUFF_LENGTH) {
        float temp = hdc2080_get_temp();
        float humid = hdc2080_get_humid();
        uint16_t record_size = 1 + sizeof(env_monitor_timestamp) + sizeof(temp) + sizeof(humid);

        // Record header - what will be written to data_buffer
        uint8_t record_header = 0x0C;
        if (measure_voc) {
            record_header |= MEASURE_VOC_MASK;
            record_size += sizeof(voc_index);
        }
        if (measure_co2) {
            record_header |= MEASURE_CO2_MASK;
            record_size += sizeof(co2_ppm);
        }

        // Write to buffer only when there is space for new data
        if ((data_buff_index + record_size) <= DATA_BUFF_LENGTH) {
            data_buff[data_buff_index] = (0xF0 | record_header);

            // Add timestamp
            data_buff_index +=1;
            memcpy(&data_buff[data_buff_index], (uint8_t *)&env_monitor_timestamp, sizeof(env_monitor_timestamp));
            data_buff_index += sizeof(env_monitor_timestamp);

            // Add temperature
            memcpy(&data_buff[data_buff_index], (uint8_t *)&temp, sizeof(temp));
            data_buff_index += sizeof(temp);

            // Add humidity
            memcpy(&data_buff[data_buff_index], (uint8_t *)&humid, sizeof(humid));
            data_buff_index += sizeof(humid);

            // Add VOC index
            if (measure_voc) {
                memcpy(&data_buff[data_buff_index], (uint8_t *)&voc_index, sizeof(voc_index));
                data_buff_index += sizeof(voc_index);
            }

            // Add CO2
            if (measure_co2) {
                memcpy(&data_buff[data_buff_index], (uint8_t *)&co2_ppm, sizeof(co2_ppm));
                data_buff_index += sizeof(co2_ppm);
            }
        }
    }
}

static firestore_err_t env_monitor_send_batt_state_change() {
    // Create document json object
    cJSON *document_json = cJSON_CreateObject();
    cJSON *fields_json = cJSON_CreateObject();
    cJSON_AddItemToObject(document_json, "fields", fields_json);

    // Add battery state to json
    cJSON *batt_state_json = cJSON_CreateObject();
    cJSON_AddBoolToObject(batt_state_json, "booleanValue", batt_adc_is_batt_empty());
    cJSON_AddItemToObject(fields_json, "empty", batt_state_json);

    char *document = cJSON_Print(document_json);

    firestore_err_t ret = com_update_document("batteryState", mac_id, document);

    cJSON_Delete(document_json);

    batt_state_sync = (ret == FIRESTORE_OK);

    return ret;
}

void env_monitor_check_battery() {
    if (!batt_adc_is_initialized()) {
        batt_adc_init();
    }

    if (batt_adc_measure() == ESP_OK) {
        if (batt_adc_check_batt_state() || !batt_state_sync) {
            env_monitor_send_batt_state_change();
        }
    }

    next_battery_measure_time.tv_sec += BATT_MEASURE_TIME;
}

esp_err_t env_monitor_start_diagnostics() {
    esp_err_t ret = ESP_OK;

    // RGB LED
    rgb_light_ms(5, 5, 5, 500);

    // UART
    uart_config_t uart_config = {
        .baud_rate = 115200,
        .data_bits = UART_DATA_8_BITS,
        .parity    = UART_PARITY_DISABLE,
        .stop_bits = UART_STOP_BITS_1,
        .flow_ctrl = UART_HW_FLOWCTRL_DISABLE
    };
    uart_param_config(UART_NUM_0, &uart_config);
    uart_set_pin(UART_NUM_0, GPIO_NUM_1, GPIO_NUM_3, UART_PIN_NO_CHANGE, UART_PIN_NO_CHANGE);
    uart_driver_install(UART_NUM_0, 1024 * 2, 0, 0, NULL, 0);
    printf("UART is driver installed: %s\n", uart_is_driver_installed(UART_NUM_0) ? "true" : "false");

    uint8_t data[4] = {0x0, 0x0, 0x0, 0x0};
    char uart_test[4] = {'a', 'b', 'c', 'd'};
    vTaskDelay(2000/portTICK_PERIOD_MS);
    uart_flush(UART_NUM_0);
    uart_write_bytes(UART_NUM_0, uart_test, 4);
    uart_read_bytes(UART_NUM_0, data, 4, 50);

    printf("Received: %c %c %c %c\n", data[0], data[1], data[2], data[3]);

    if ((data[0] != uart_test[0]) || (data[1] != uart_test[1]) || (data[2] != uart_test[2]) || (data[3] != uart_test[3])) {
        rgb_light_ms(5, 0, 0, 500);
    }

    // ADC
    if (!batt_adc_is_initialized()) {
        batt_adc_init();
    }

    if (batt_adc_start_diagnostics() == ESP_OK) {
        printf("ADC init successful\n");
    }
    else {
        ESP_LOGE(TAG, "ADC init error");
    }

    // Init I2C0 for HDC2080 and SGP40
    esp_err_t i2c_ret = sens_i2c0_init();

    // HDC2080
    if (i2c_ret == ESP_OK) {
        uint8_t data[2];
        esp_err_t hdc_ret = sens_i2c_read_from_reg(I2C_NUM_0,  HDC2080_ADDR, HDC2080_MANUFACTURER_ID_L_REG, data, sizeof(data));
        if (hdc_ret == ESP_OK) {
            printf("HDC2080 init successful, manufacturer id:  %x%x\n", data[0], data[1]);
        }
        else {
            ESP_LOGE(TAG, "HDC2080 init error");
            ret = ESP_FAIL;
        }
    }

    // SGP40
    if (i2c_ret == ESP_OK) {
        uint16_t timeout = 0;
        while ((sgp40_probe() != STATUS_OK) && (timeout < 100)) {
            timeout++;
            vTaskDelay((1/portTICK_PERIOD_MS) + 1);
        }
        if (timeout < 100) {
            printf("SGP sensor probing successful\n");
        }
        else {
            ESP_LOGE(TAG, "SGP40 init error");
            ret = ESP_FAIL;
        }
    }

    if (ret == ESP_OK) {
        rgb_light_ms(2, 0, 0, 500);
        rgb_light_ms(0, 2, 0, 500);
        rgb_light_ms(0, 0, 2, 500);
        rgb_light_ms(2, 2, 2, 500);
    }
    else {
        rgb_light_ms(2, 0, 0, 500);
    }

    return ret;
}

esp_err_t env_monitor_enable_tps_power_save_mode() {
    esp_err_t ret;

    ret = gpio_set_level(TPS_PS_EN_GPIO, 0);

    if (ret == ESP_OK) {
        printf("Power save mode enabled\n");
    }
    else {
        ESP_LOGE(TAG, "Power save mode enable error\n");
    }

    return ret;
}

esp_err_t env_monitor_disable_tps_power_save_mode() {
    esp_err_t ret;

    ret = gpio_set_level(TPS_PS_EN_GPIO, 1);

    if (ret == ESP_OK) {
        printf("Power save mode disabled\n");
    }
    else {
        ESP_LOGE(TAG, "Power save mode disable error\n");
    }
    
    return ret;
}