#include "string.h"
#include "stdio.h"
#include "stdlib.h"
#include "esp_log.h"
#include "nvs_flash.h"
#include "esp_event.h"
#include "freertos/event_groups.h"
#include "esp_wifi.h"
#include "driver/i2c.h"
#include "esp_sleep.h"

#include "firestore.h"
#include "cJSON.h"

#include "battery_adc.h"
#include "communication.h"
#include "env_monitor.h"
#include "rgb.h"
#include "sensors.h"
#include "sntp_time.h"

#define TAG "MAIN"

void app_main() {
    esp_log_level_set("wifi", ESP_LOG_WARN);

    // Initialize NVS
    esp_err_t ret = nvs_flash_init();
    if (ret == ESP_ERR_NVS_NO_FREE_PAGES || ret == ESP_ERR_NVS_NEW_VERSION_FOUND)
    {
        ESP_ERROR_CHECK(nvs_flash_erase());
        ret = nvs_flash_init();
    }
    ESP_ERROR_CHECK(ret);

    // Initialize PWM for RGB LED
    rgb_pwm_init();

    // Deep sleep - get wakeup cause
    env_monitor_wakeup_t wakeup_cause = env_monitor_get_wakeup_cause();

    bool init_time = false;
    bool load_settings = false;

    // Set default settings and process diagnostics after start
    if (!env_monitor_is_initialized() || (wakeup_cause == ENV_MONITOR_WAKEUP_UNDEFINED)) {
        if (!env_monitor_is_initialized()) {
            env_monitor_voc_init();
        }

        env_monitor_start_diagnostics();
        env_monitor_get_mac_id();
        env_monitor_set_default_settings();

        // Init TPS_PS_EN pin and enable power save mode
        gpio_set_direction(TPS_PS_EN_GPIO, GPIO_MODE_OUTPUT);
        env_monitor_enable_tps_power_save_mode();

        init_time = true;
        load_settings = true;
    }

    // Get current time from SNTP server
    if (init_time || env_monitor_get_sntp_retry_next() || env_monitor_get_sntp_sync_time_flag()) {
        if (com_wifi_init_sta() == ESP_OK) {
            esp_err_t sntp_ret = sntp_time_get_current_time();
            if (sntp_ret != ESP_OK) {
                env_monitor_set_sntp_retry_next(true);
                env_monitor_sleep_start();
            }
            else {
                env_monitor_set_sntp_retry_next(false);
            }
        }
        env_monitor_set_default_tasks_time();
    }

    // Load settings from Firestore
    if (load_settings || env_monitor_get_load_settings_retry_next()) {
        if (com_is_wifi_initialized() || (com_wifi_init_sta() == ESP_OK)) {
            firestore_init();
            if (env_monitor_load_settings() != FIRESTORE_OK) {
                env_monitor_set_load_settings_retry_next(true);
            }
            else {
                env_monitor_set_load_settings_retry_next(false);
            }
        }
    }

    switch (wakeup_cause) {
        case ENV_MONITOR_WAKEUP_TIMER:
        case ENV_MONITOR_WAKEUP_UNDEFINED: {
            // Check tasks to do
            env_monitor_check_tasks();

            // Measure battery voltage
            if (env_monitor_battery_measure_task_ticked()) {
                if (com_is_wifi_initialized() || (com_wifi_init_sta() == ESP_OK)) {
                    firestore_init();
                    env_monitor_check_battery();
                }
            }

            // Measure data
            if (env_monitor_measure_task_ticked()) {
                if ((sens_get_i2c0_initialized() || (sens_i2c0_init() == ESP_OK)) && (sens_get_i2c1_initialized() || (sens_i2c1_init() == ESP_OK))) {
                    
                    esp_err_t retMeasure = env_monitor_measure_data();

                    if (!env_monitor_send_task_ticked() && (retMeasure == ESP_OK)) {
                        env_monitor_add_data_to_buff();
                    }
                }
            }

            // Process VOC algorithm
            if (env_monitor_voc_algorithm_task_ticked()) {
                if (sens_get_i2c0_initialized() || (sens_i2c0_init() == ESP_OK)) {
                    env_monitor_process_voc_algorithm();
                }
            }

            // Send data
            if (env_monitor_send_task_ticked()) {
                if (com_is_wifi_initialized() || (com_wifi_init_sta() == ESP_OK)) {
                    firestore_init();
                    env_monitor_send_data();
                }
            }
            break;
        }
        case ENV_MONITOR_WAKEUP_GPIO34: {
            // Get parameters from Firestore
            if (com_is_wifi_initialized() || (com_wifi_init_sta() == ESP_OK)) {
                firestore_init();
                if (env_monitor_load_settings() != FIRESTORE_OK) {
                    env_monitor_set_load_settings_retry_next(true);
                }
                else {
                    env_monitor_set_load_settings_retry_next(false);
                }
            }
            break;
        }
        case ENV_MONITOR_WAKEUP_GPIO35: {
            // Zde pripadne pribude dalsi rozsireni
            rgb_light_ms(5, 5, 5, 500);
        }
        default: {
            break;
        }
    }

    env_monitor_sleep_start();
}